/*--------------------------------------------------------------------------
@(#)tcpip.h	1.6 12/2/92

 The Berkeley Sockets programming interface to the TCP/IP protocol suite
 is widely supported, but the exact names of the entry points change
 from implementation to implementation.

 This file uses #define to provide a consistant set of entry point names:
 netread(), netwrite(), socket(), netioctl(), select(), and
 netperror(), as well as a consistantly named global, int neterrno.

 Some semantic differences persist in poorer implementations; for instance,
 gethostbyname() doesn't work properly under Wollengong,
 and select() doesn't work with all versions.

 The original version of this file was donated by Mike Kienenberger 
 <FSMLK1@acad3.alaska.edu>.  Please forward any corrections to him, and 
 to the uwho group at uwho@blacks.jpl.nasa.gov.
--------------------------------------------------------------------------*/

#ifdef VMS
/* vms cc can't /define more than one thing... so options go in .h file */
#include "vmsdef.h"
#endif

#ifdef WOLLENGONG
/* VAX/VMS Wollengong */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>	/* for inet_addr */
#include <sys/ioctl.h>          /* for FIONBIO */
#include <errno.h>              /* for EINPROGRESS */
#define USE_FIONBIO             /* this system uses ioctl to set nonblocking.*/
#define netioctl ioctl
#define neterrno errno
#define netperror perror

/* End of WOLLENGONG def */
#endif
#ifdef MULTINET
/* VAX/VMS TGV Multinet */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <multi/netdb.h>
#include <multi/errno.h>	/* for EINPROGRESS */
#include <sys/ioctl.h>		/* for FIONBIO */
#include <netinet/in.h>
#include <arpa/inet.h>	/* for inet_addr */

#define netclose socket_close
#define netread socket_read
#define netwrite socket_write
#define netioctl socket_ioctl
#define neterrno socket_errno
#define netperror socket_perror
#define USE_FIONBIO		/* this system uses ioctl to set nonblocking.*/

/* End of MULTINET def */
#endif
#ifdef UCX
/* VAX/VMS DEC UCX aka TCP/IP services for VMS */

#include <types.h>
#include <sys/time.h>
#include <socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>	/* for inet_addr */

#define netclose close
#define netread read
#define netwrite write
#define netioctl ioctl
#define neterrno errno
#define netperror perror

/* End of UCX def */
#endif
#ifdef NOVELL_LWP

/* Experimental Novell tcp/ip definitions. */
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>	/* for inet_addr */
#include <sys/filio.h>          /* for FIONBIO */
#define USE_FIONBIO             /* this system uses ioctl to set nonblocking.*/

#define netclose soclose
#define netread soread
#define netwrite sowrite
#define netioctl ioctl
#define neterrno errno          /* really? */
#define netperror soperror
 
/* End of NOVELL_LWP def */

#endif

#ifdef WINSOCK
#include <winsock.h>
/* defines later */
#endif

#ifdef UNIX
/* Unix */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>	/* for inet_addr */

#ifdef NEXT
/* most of unistd.h is in libc.h   */
/* most of malloc.h is in stdlib.h */
#else
#include <unistd.h>
#include <malloc.h>
#endif

/* FNDELAY and O_NDELAY: */
#include <fcntl.h>
#ifdef HPUX
/* HP/UX defines FNDELAY in sys/file.h which is not included */
#define USE_O_NDELAY
#endif

#define netclose close
#define netread read
#define netwrite write
#define netioctl ioctl
#define neterrno errno
#define netperror perror

/* End of UNIX def */
#endif
